/*------------------------------------------------------------------------
GridActor.cpp: implementation of the CGridActor class.
GridSlammer Game Engine - version 0.1

    Copyright (C) 2000 Thad D. Phetteplace

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Examine the license.txt file located in the GridSlammer directory
    to see the full text of the LGPL license document.

    Direct inquiries to:    The GridSlammer Project
                            P.O. Box 26354
                            Wauwatosa, WI 53208

    Or on the Internet at:    http://www.gridslammer.org
                            info@gridslammer.org
------------------------------------------------------------------------*/

#include "GridActor.h"
#include <stdio.h>

//-------------------------------------------------------------------------
// Method:    CGridActor
// Paramters: none
//
// Returns:   nothing
//
// The constructor.
//-------------------------------------------------------------------------
CGridActor::CGridActor()
{
    Initialize();
}




//-------------------------------------------------------------------------
// Method:    CGridActor
// Paramters: Info = pointer to a a CGridInfo structure
//
// Returns:   nothing
//
// This version of the constructor instantiates the grid object and then
// connects it to the game grid by calling the ConnectToGrid method.
//-------------------------------------------------------------------------
CGridActor::CGridActor(GGridInfo *Info)
{
    Initialize();
    ConnectToGrid(Info);
}



//-------------------------------------------------------------------------
// Method:    Initialize
// Paramters: none
//
// Returns:   nothing
//
// Initializes various member variables to default values.
//-------------------------------------------------------------------------
void CGridActor::Initialize( void )
{
    // Initialize the base class stuff
    CGridObj::Initialize();

    // Initialize our stuff
    m_Views = 1;
    m_Speed = 0.0;
    m_XSpeed = 0.0;
    m_YSpeed = 0.0;
    m_Direction = 0.0;
}



//-------------------------------------------------------------------------
// Method:    ~CGridActor
// Paramters: none
//
// Returns:   nothing
//
// The destructor.
//-------------------------------------------------------------------------
CGridActor::~CGridActor()
{
}





//-------------------------------------------------------------------------
// Method:    Render
// Paramters: gv = pointer to the GridView to render into
//            lClockTicks = the number of milliseconds since game start
//            ScreenX, nScreenY = the screen coordinates to draw to
//
// Returns:   Zero on success, non zero on failure
//
// Draws the grid object to the screen at the supplied coordinates.
// This function is called by the framework as part of the overall
// rendering of the game.
//
// Override this method with your own code if you want to do something
// other than draw a static image.
//-------------------------------------------------------------------------
int CGridActor::Render(CGridView *gv, unsigned long lClockTicks,
    int nScreenX, int nScreenY)
{
    CGridObj::Render(gv, lClockTicks, nScreenX, nScreenY);
    return 0;
}





//-------------------------------------------------------------------------
// Method:    Process
// Paramters: lClockTicks = the number of milliseconds since game start
//            dTimeSlice = the unit of game time to process
//
// Returns:   Zero on success, non zero on failure
//
// Process a single iteration of game action for this object.  The
// dTimeSlice parameter should be used to select the amount of action
// to perform; it is dynamically tuned by the framework to a adjust for
// individual for individual system's performance and rendering framerate.
// This function is called periodically by the framework in between
// rendering of frames as part of the overall game simulation.
//
// The GridActor version of the Process method does considerably more
// than the GridObj version.  It adjusts the position of the object within
// the grid and performs collision detection with tiles and other
// objects.
//-------------------------------------------------------------------------
int CGridActor::Process(unsigned long lClockTicks, double dTimeSlice)
{
    double newx, newy;
    CGridObj* p;

    // Get the new X and Y position
    newx = m_XPos + m_XSpeed * dTimeSlice;
    newy = m_YPos + m_YSpeed * dTimeSlice;

    // Make sure we did not collide with anything
    int status =  CollisionCheck(newx, newy, &p);
    if ( 1 == status )
    { //we have collided with a wall, we will just  check along horizontal
      //and vertical lines of movement
      newx = m_XPos;//move in direction of y-axis
        if ( CollisionCheck(newx, newy, &p) )
        { // Nope, reset Y and try X only
            newx = m_XPos + m_XSpeed * dTimeSlice;
            newy = m_YPos;
            if ( CollisionCheck(newx, newy, &p) )
            {// Can't move either way, reset both X and Y
                newx = m_XPos;
                newy = m_YPos;
            }
        }
    }
    else if ( 2 == status )
    { //we have collided with object p, try if we can get by in a way
      //consistent with our movement direction
      double xTarget, yTarget, factor, modXspeed, modYspeed;
      xTarget = (p->m_XPos) - m_XPos;
      yTarget = (p->m_YPos) - m_YPos;
      //get the component of the speed vector that is at right angles to the
      //target vector
      factor = dTimeSlice * (xTarget * m_XSpeed + yTarget * m_YSpeed) / 
               ( xTarget * xTarget + yTarget * yTarget );
      modXspeed = m_XSpeed * dTimeSlice - factor *  xTarget;
      modYspeed = m_YSpeed * dTimeSlice - factor *  yTarget;
      //now try going in that direction
      newx = m_XPos + modXspeed;
      newy = m_YPos + modYspeed;
      if ( CollisionCheck(newx , newy , &p) ) //we collide at this location also
      { //give up trying to move
          newx = m_XPos;
          newy = m_YPos;
      } //else we move to the new position computed here
    }

    // Remove the old grid reference
    RemoveObjFromSquare();

    // don't let it go off the edge
    // This should never happen in a properly built grid.
    if (newx < 0.0) newx = 0.0;
    else if (newx > (float) MAX_X_GRID_SIZE - 0.1) newx = 99.9;
    m_XPos = newx;
    if (newy < 0.0) newy = 0.0;
    else if (newy > (float) MAX_Y_GRID_SIZE - 0.1) newy = 99.9;
    m_YPos = newy;

    // Create a new grid reference to the object
    AddObjToSquare();

    return 0;
}




//-------------------------------------------------------------------------
// Method:    ChangeDirection
// Paramters: change = amount to change direction (measured in radians)
//
// Returns:   nothing
//
// Turn the object so it is facing (and probably moving) in another
// direction.  The change parameter can be passed in as possative or
// negative radian measure to cause clockwise or counterclockwise
// rotation.
//-------------------------------------------------------------------------
void CGridActor::ChangeDirection( double change )
{

	// Adjust the speed
    m_Direction += change;

	// Make sure we haven't looped all the way around
    if (m_Direction < 0.0)
        m_Direction += TWO_PI_RADIANS;

    if (m_Direction > TWO_PI_RADIANS)
        m_Direction -= TWO_PI_RADIANS;

	// Recalculate the X and Y components of the speed
    m_XSpeed = CalcXSpeed( m_Speed, m_Direction );
    m_YSpeed = CalcYSpeed( m_Speed, m_Direction );
}



//-------------------------------------------------------------------------
// Method:    ChangeSpeed
// Paramters: change = amount to change direction (measured in radians)
//
// Returns:   nothing
//
// Change the overall speed of the object.
//-------------------------------------------------------------------------
void CGridActor::ChangeSpeed( double change )
{
    m_Speed += change;
	// Change the X and Y components of the speed simply by calling
	// the ChangeDirection method with a zero direction change (speed
	// is always recalculated when direction changes)
    ChangeDirection( 0.0 );
}


