/*------------------------------------------------------------------------
GridObj.cpp: implementation of the CGridObj class.
GridSlammer Game Engine - version 0.1

    Copyright (C) 2000 Thad D. Phetteplace

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Examine the license.txt file located in the GridSlammer directory
    to see the full text of the LGPL license document.

    Direct inquiries to:    The GridSlammer Project
                            P.O. Box 26354
                            Wauwatosa, WI 53208

    Or on the Internet at:  http://www.gridslammer.org
                            info@gridslammer.org
------------------------------------------------------------------------*/

#include "GridObj.h"


//-------------------------------------------------------------------------
// Method:    CGridObj
// Paramters: none
//
// Returns:   nothing
//
// The constructor.
//-------------------------------------------------------------------------
CGridObj::CGridObj()
{
    Initialize();
}




//-------------------------------------------------------------------------
// Method:    CGridObj
// Paramters: Info = pointer to a a CGridInfo structure
//
// Returns:   nothing
//
// This version of the constructor instantiates the grid object and then
// connects it to the game grid by calling the ConnectToGrid method.
//-------------------------------------------------------------------------
CGridObj::CGridObj(GGridInfo *Info)
{
    Initialize();
    ConnectToGrid(Info);
}




//-------------------------------------------------------------------------
// Method:    ConnectToGrid
// Paramters: Info = pointer to a a CGridInfo structure
//
// Returns:   nothing
//
// This method assigns various member variables, including pointers into
// member variables with the GridStage.
//-------------------------------------------------------------------------
void CGridObj::ConnectToGrid(GGridInfo *Info)
{
    m_Grid = Info->Grid;
    m_GridObjs = (CGridObj **) Info->GridObjs;
    m_Tiles = (CGridTile **) Info->Tiles;
    m_Objects = (CGridObj **) Info->Objects;
    m_GridXSize = MAX_X_GRID_SIZE;  // this should be dynamic! -TDP
    m_pTotalGridObjs = Info->TotalGridObjs;
    m_ObjNum = *m_pTotalGridObjs;
    m_Objects[(*m_pTotalGridObjs)++] = this;  // Range check this! -TDP
    m_AudioQueue = Info->AudioQueue;
    m_Sounds = Info->Sounds;
    //AddObjToSquare();
}




//-------------------------------------------------------------------------
// Method:    Initialize
// Paramters: none
//
// Returns:   nothing
//
// Initializes various member variables to default values.
//-------------------------------------------------------------------------
void CGridObj::Initialize( void )
{
    m_XPos = 50.0;
    m_YPos = 50.0;
    m_XOffset = 0;
    m_YOffset = 0;
    m_Diameter = 1.0;
    m_Radius = 0.5;
    m_PicNum = 2;
    m_Solid = 1;
    m_Team = 0;
    m_ObjType = 0;  // set this?  -TDP
    m_NextObj = (CGridObj *) 0;
}





//-------------------------------------------------------------------------
// Method:    ~CGridObj
// Paramters: none
//
// Returns:   nothing
//
// The destructor.
//-------------------------------------------------------------------------
CGridObj::~CGridObj()
{
}




//-------------------------------------------------------------------------
// Method:    Render
// Paramters: gv = pointer to the GridView to render into
//            lClockTicks = the number of milliseconds since game start
//            ScreenX, nScreenY = the screen coordinates to draw to
//
// Returns:   Zero on success, non zero on failure
//
// Draws the grid object to the screen at the supplied coordinates.
// This function is called by the framework as part of the overall
// rendering of the game.
//
// Override this method with your own code if you want to do something
// other than draw a static image.
//-------------------------------------------------------------------------
int CGridObj::Render(CGridView *gv, unsigned long lClockTicks,
             int nScreenX, int nScreenY)
{
    if (IsSelected()) {
        gv->DrawPic( m_PicNum, nScreenX+m_XOffset, nScreenY+m_YOffset,
             DP_HILIGHT);
    }
    else {
        gv->DrawPic( m_PicNum, nScreenX+m_XOffset, nScreenY+m_YOffset);
    }

    return 0;
}





//-------------------------------------------------------------------------
// Method:    Process
// Paramters: lClockTicks = the number of milliseconds since game start
//            dTimeSlice = the unit of game time to process
//
// Returns:   Zero on success, non zero on failure
//
// Process a single iteration of game action for this object.  The
// dTimeSlice parameter should be used to select the amount of action
// to perform; it is dynamically tuned by the framework to a adjust for
// individual for individual system's performance and rendering framerate.
// This function is called periodically by the framework in between
// rendering of frames as part of the overall game simulation.
//-------------------------------------------------------------------------
int CGridObj::Process(unsigned long lClockTicks, double dTimeSlice)
{
    RemoveObjFromSquare();
    AddObjToSquare();
    return 0;
}





//-------------------------------------------------------------------------
// Method:    AddObjToSquare
// Paramters: none
//
// Returns:   none
//
// This function adds the current object to the linked list of grid
// objects associated with the square it is occupying.  It needs to be
// called whenever an object enters a square.  This is usually done for
// you in the Process method.
//-------------------------------------------------------------------------
void CGridObj::AddObjToSquare( void )
{
    CGridObj *p;
    CGridObj *prev;
    double x, y, px, py;

    // get the first object in the square
    p = GRIDOBJS(m_XPos, m_YPos);

    // if the square is empty, just stuff the object in the first space
    if ( p == 0 ) {
        m_NextObj = 0;
        GRIDOBJS(m_XPos, m_YPos) = this;
        return;
    }

    // the square is not empty, so compute if our object is closer
    // than the one in the head of the list.
    x = m_XPos - (double) ((int) m_XPos);
     y = m_YPos - (double) ((int) m_YPos);
    px = p->m_XPos - (double) ((int) p->m_XPos);
     py = p->m_YPos - (double) ((int) p->m_YPos);

    if (x*x + y*y >= px*px + py*py) {
        // our object is closer, so it becomes the head of the list
        m_NextObj = p;
        GRIDOBJS(m_XPos, m_YPos) = this;
        return;
    }

    // loop until we find the insertion spot
    do {
        // move to the next object in the list
        prev = p;
        p = p->m_NextObj;

        if (p) {
            x = m_XPos - (double) ((int) m_XPos);
            y = m_YPos - (double) ((int) m_YPos);
            px = p->m_XPos - (double) ((int) p->m_XPos);
            py = p->m_YPos - (double) ((int) p->m_YPos);
        }
    } while ( p && (x*x + y*y < px*px + py*py) );

    // insert the object into the list
    m_NextObj = p;
    prev->m_NextObj = this;

    // Cool! we are done!
}





//-------------------------------------------------------------------------
// Method:    RemoveObjFromSquare
// Paramters: none
//
// Returns:   none
//
// This function removes the current object from the linked list of
// grid objects associated with the grid square it is occupying.  This
// function should be called whenever an object leaves a square.  This
// is usually done for you in the Process method.
//-------------------------------------------------------------------------
void CGridObj::RemoveObjFromSquare( void )
{
    CGridObj *p;
    CGridObj *prev;
    int x;
    int y;

    x = (int) m_XPos;
    y = (int) m_YPos;

    if ( GRIDOBJS(x,y) == this ) {
        GRIDOBJS(x,y) = GRIDOBJS(x,y)->m_NextObj;
    }
    else if (GRIDOBJS(x,y)) {
        prev = GRIDOBJS(x,y);
        p = prev->m_NextObj;
        while ( p && p != this ) {
            prev = p;
            p = p->m_NextObj;
        }
        if ( p == this ) {
            prev->m_NextObj = p->m_NextObj;
        }
    }
}





//-------------------------------------------------------------------------
// Method:    DestroyObject
// Paramters: none
//
// Returns:   nothing
//
// Remove this grid object from the list of active objects held by the
// framework and then delete the object itself.  Call this object when
// an game object is consumed or a creature is destroyed.
//-------------------------------------------------------------------------
void CGridObj::DestroyObject( void )
{
    RemoveObjFromSquare();
    (*m_pTotalGridObjs)--;
    m_Objects[m_ObjNum] = m_Objects[*m_pTotalGridObjs];
    m_Objects[m_ObjNum]->m_ObjNum = m_ObjNum;
    delete this;
}





//-------------------------------------------------------------------------
// Method:    CollisionCheck
// Paramters: nXpos, nYpos = the grid position to check for a collision
//            pObj = Used to return a pointer to the object that we
//                   collided with.
//
// Returns:   0 = no collision
//            1 = collided with a tile
//            2 = collided with an object
//
// Call this function with a specific grid coordinate, and it will return
// if that position collides with a tile or grid object.  If it collides
// with a grid object, it returns a pointer to that object in the pObj
// parameter.
//-------------------------------------------------------------------------
int CGridObj::CollisionCheck( double nXpos, double nYpos, CGridObj **pObj )
{
    *pObj = 0;  // return a null object pointer by default

	// Check if we collide with a tile
    if ( TILE(nXpos, nYpos)->CollisionCheck(nXpos, nYpos)
      || TILE(nXpos+m_Diameter, nYpos)->CollisionCheck(
              nXpos+m_Diameter, nYpos)
      || TILE(nXpos, nYpos+m_Diameter)->CollisionCheck(
              nXpos, nYpos+m_Diameter)
      || TILE(nXpos+m_Diameter, nYpos+m_Diameter)->CollisionCheck(
              nXpos+m_Diameter, nYpos+m_Diameter) ) {
          return 1;
    }

	// Check if we collide with an object
    if ( (*pObj = ObjCollisionCheck(nXpos, nYpos, nXpos, nYpos)) ||
         (*pObj = ObjCollisionCheck(nXpos, nYpos, nXpos+m_Diameter, nYpos)) ||
         (*pObj = ObjCollisionCheck(nXpos, nYpos, nXpos, nYpos+m_Diameter)) ||
         (*pObj = ObjCollisionCheck(nXpos, nYpos, nXpos+m_Diameter, nYpos+m_Diameter)) ||
         (*pObj = ObjCollisionCheck(nXpos, nYpos, nXpos+m_Diameter, nYpos-m_Diameter)) ||
         (*pObj = ObjCollisionCheck(nXpos, nYpos, nXpos-m_Diameter, nYpos+m_Diameter)) ||
         (*pObj = ObjCollisionCheck(nXpos, nYpos, nXpos-m_Diameter, nYpos)) ||
         (*pObj = ObjCollisionCheck(nXpos, nYpos, nXpos, nYpos-m_Diameter)) ||
         (*pObj = ObjCollisionCheck(nXpos, nYpos, nXpos-m_Diameter, nYpos-m_Diameter))) {
         return 2;
    }

    return 0;  // there was no collision, return zero
}





//-------------------------------------------------------------------------
// Method:    ObjCollisionCheck
// Paramters: nXpos, nYpos = The grid coordinates to check for collision
//            nXcorner, nYcorner = Used to locate the accual square to
//                                 search for grid objects.
//
// Returns:   A pointer to a grid object or null of there is no collision.
//
// This function is called by the CollisionCheck function as part of the
// overall collision detection process.  We must supply the nXcorner and
// nYcorner paramters because a grid object can span more then one square,
// so we must search the various squares around us to chech for collisions.
// An object is only linked in one square at a time, even though it can
// span more than one square; thus we must check in more than just the
// square we are in.
//-------------------------------------------------------------------------
CGridObj* CGridObj::ObjCollisionCheck( double nXpos, double nYpos, double nXcorner, double nYcorner )
{
    CGridObj* p;
    double x, y;
    double r;

	// Locate the grid square to look in
    p = GRIDOBJS((int) nXcorner, (int) nYcorner);

	// Loop through the linked list of objects in this square
    while (p) {
        if ( p != this ) {
            x = nXpos - p->m_XPos;  // get the difference in x distance
            y = nYpos - p->m_YPos;  // get the difference in y distance
            r = m_Radius + p->m_Radius;  // Add up the radius of the two objects
			// Now check if the two bounding circles are intersecting.
			// Note that we are basically using the standard distance formula,
			// but with both sides squared to remove the costly square root
			// computation.
            if ( p->m_Solid && (x*x + y*y < r*r) ) {
                return p;  // return the pointer to the object
            }
        }
        p = p->m_NextObj;  // get the next object in the list
    }

    return 0;  // there was no collision, return zero
}





//-------------------------------------------------------------------------
// Method:    PlaySound
// Paramters: SoundNum = the index number of the sound to play
//
// Returns:   zero on success, non zero on failure
//
// This function queues up a sound to be played by the framework.  There
// is a limit to the number of sounds that can play at a time, so it is
// possible for this function to fail if there is a lot going on in the
// game.
//-------------------------------------------------------------------------
int CGridObj::PlaySound( int SoundNum )
{
    int i;

    for (i=0; i<MAX_AUDIO_EVENTS; i++) {
        if (m_AudioQueue[i].len == 0) {
            m_AudioQueue[i].len = m_Sounds[SoundNum].len;
            m_AudioQueue[i].pos = m_Sounds[SoundNum].buf;
            return 0;
        }
    }

    return 1;
}




//-------------------------------------------------------------------------
// Method:    FillGridInfo
// Paramters: Info = a pointer to the GridInfo structure to fill
//
// Returns:   A pointer to the Info structure that was passed in.
//
// Call this function to fill a GridInfo structure with information from
// the current grid object.  This saves you the effort of assigning all
// the fields manually, and is particularly useful when your grid object
// is spawning new objects (such as bullets) and connecting them to the
// grid.
//-------------------------------------------------------------------------
GGridInfo* CGridObj::FillGridInfo(GGridInfo *Info)
{
    Info->Grid = m_Grid;
    Info->GridObjs = m_GridObjs;
    Info->Tiles = m_Tiles;
    Info->Objects = m_Objects;
    Info->TotalGridObjs = m_pTotalGridObjs;
    Info->Sounds = m_Sounds;
    Info->AudioQueue = m_AudioQueue;

    return Info;
}


